if Code.ensure_loaded?(NimbleCSV) do
  defmodule ClickHouse.Format.TSV do
    @moduledoc """
    An implementation of the ClickHouse `"TSV"`/`"TabSeparatedValues"` format.
    """

    @behaviour ClickHouse.Format

    alias ClickHouse.{DataType, Query}

    @parser __MODULE__.Parser

    NimbleCSV.define(@parser, moduledoc: false, separator: "\t", escape: "\"")

    ################################
    # ClickHouse.Format Callbacks
    ################################

    @impl ClickHouse.Format
    @spec names() :: [binary()]
    def names, do: ["TabSeparated", "TSV"]

    @impl ClickHouse.Format
    @spec decode(ClickHouse.Result.t()) :: ClickHouse.Result.t()
    def decode(result), do: do_decode(result)

    @impl ClickHouse.Format
    @spec encode(ClickHouse.Query.t()) :: ClickHouse.Query.t()
    def encode(%Query{command: :insert, params_count: count}) when count > 1 do
      raise ArgumentError, "TSV insert statement can only support a single parameter."
    end

    def encode(%Query{command: :insert, params_count: 0, params: [_ | _]}) do
      raise ArgumentError, "TSV insert statement with params must include a '$1'."
    end

    def encode(%Query{command: :insert, params_count: 1, params: params} = query) do
      params = do_encode(params)
      statement = String.trim_trailing(query.statement, "$1")
      Query.put_encoded(query, [statement, params])
    end

    ################################
    # Private API
    ################################

    defp do_decode(result) do
      data = @parser.parse_string(result.raw, skip_headers: false)
      %{result | data: data}
    end

    defp do_encode(params) do
      params
      |> Enum.map(&encode_row/1)
      |> @parser.dump_to_iodata()
    end

    defp encode_row(row), do: Enum.map(row, &encode_value/1)

    defp encode_value(nil), do: nil
    defp encode_value(data), do: DataType.encode(data)
  end
end
